import arrow
from typing import TypedDict
from enum import Enum, auto


# 采用 ErrMsg 而不是采用 exception, 一来是受到 Go 语言的影响，
# 另一方面，凡是用到 ErrMsg 的地方都是与业务逻辑密切相关并且需要向用户反馈详细错误信息的地方，
# 这些地方用 ErrMsg 更合理。 (以后会改用 pypi.org/project/result)
ErrMsg = str
"""一个描述错误内容的简单字符串，空字符串表示无错误。"""

NotFound: ErrMsg = "NotFound"


def now() -> float:
    return arrow.now().timestamp()


class TodoStatus(Enum):
    Incomplete = auto()
    Completed = auto()


class Repeat(Enum):
    Never = auto()
    Week = auto()
    Month = auto()
    Year = auto()


class TodoItem(TypedDict):
    ctime: float  # create-time, 用于排序, 同时也当作 ID
    dtime: float  # done-time, 完成时间，只用于排序
    event: str
    status: str  # TodoStatus
    repeat: str  # Repeat
    s_date: str  # start-date, 第一次提醒日期, "YYYY-MM-DD"
    n_date: str  # next-date, 下次提醒日期, "YYYY-MM-DD"


TodoList = list[TodoItem]
IdxTodoList = list[tuple[int, TodoItem]]  # list of (index, item)


def new_todoitem(event: str) -> TodoItem:
    return TodoItem(
        ctime=now(),
        dtime=0,
        event=event,
        status=TodoStatus.Incomplete.name,
        repeat=Repeat.Never.name,
        s_date="",
        n_date="",
    )


class DB(TypedDict):
    u_date: str  # "YYYY-DD-MM" 用来判断要不要刷新周期计划（每天刷新一次）
    items: list[TodoItem]


def new_db() -> DB:
    return DB(u_date="", items=[])



import click
import json
import arrow
from pathlib import Path
from typing import cast
from appdirs import AppDirs
from arrow.arrow import Arrow

from simpletodo.model import (
    DB,
    ErrMsg,
    IdxTodoList,
    Repeat,
    TodoItem,
    TodoList,
    TodoStatus,
    new_db,
    now,
)

todo_db_name = "todo-db.json"
DateFormat = "YYYY-MM-DD"

app_dirs = AppDirs("todo", "github-ahui2016")
app_config_dir = Path(app_dirs.user_config_dir)
db_path = app_config_dir.joinpath(todo_db_name)


def ensure_db_file() -> None:
    app_config_dir.mkdir(parents=True, exist_ok=True)
    if not db_path.exists():
        with open(db_path, "w", encoding="utf-8") as f:
            json.dump(new_db(), f, indent=4, ensure_ascii=False)


def load_db() -> DB:
    with open(db_path, "rb") as f:
        return cast(DB, json.load(f))


def split_db(db: DB) -> tuple[IdxTodoList, IdxTodoList, IdxTodoList]:
    todo_list: IdxTodoList = []
    done_list: IdxTodoList = []
    repeat_list: IdxTodoList = []
    for idx, item in enumerate(db["items"]):
        if TodoStatus[item["status"]] is TodoStatus.Incomplete:
            todo_list.append((idx, item))
        if item["dtime"] > 0:
            done_list.append((idx, item))
        if Repeat[item["repeat"]] is not Repeat.Never:
            repeat_list.append((idx, item))
    todo_list.sort(key=lambda x: x[1]["ctime"], reverse=True)
    done_list.sort(key=lambda x: x[1]["dtime"], reverse=True)
    repeat_list.sort(key=lambda x: x[1]["n_date"])
    return todo_list, done_list, repeat_list


def update_db(db: DB) -> None:
    with open(db_path, "w", encoding="utf-8") as f:
        json.dump(db, f, indent=4, ensure_ascii=False)


def print_todolist(l: IdxTodoList, all: bool) -> None:
    print("\nTodo\n------------")
    if not l:
        print("(none)")
        if not all:
            print("\nTry 'todo -a' to include completed items.")
        return
    for idx, item in l:
        print(f"{idx+1}. {item['event']}")


def print_donelist(l: IdxTodoList) -> None:
    print("\nCompleted\n------------")
    if not l:
        print("(none)")
        return
    for idx, item in l:
        print(f"{idx+1}. {item['event']}")


def print_repeatlist(l: IdxTodoList) -> None:
    print("\nSchedule\n------------")
    if not l:
        print("(none)")
        return
    for idx, item in l:
        repeat = item["repeat"]
        if Repeat[repeat] is Repeat.Week:
            print(
                f"{idx+1}. every {arrow.get(item['s_date']).format('dddd')} "
                f"[{item['n_date']}] {item['event']}"
            )
        else:
            print(f"{idx+1}. every {repeat.lower()} [{item['n_date']}] {item['event']}")


def is_last_day(date: Arrow) -> bool:
    """Is it the last day of month?"""
    last_day = date.ceil("month").format(DateFormat)
    return date == last_day


def validate_n(l: TodoList, n: int) -> ErrMsg:
    if not l:
        return "There is no item in the list."
    if n < 1:
        return "Please input a number bigger than zero."

    size = len(l)
    if n > size:
        if size == 1:
            return "There is only 1 item."
        else:
            return f"There are only {size} items"
    return ""


def stop_schedule(db: DB, idx: int, item:TodoItem, ctx:click.Context,) -> None:
    if Repeat[item["repeat"]] is Repeat.Never:
        click.echo("Warning: It is not set to repeat, nothing changes.")
        ctx.exit()
    db["items"][idx]["repeat"] = Repeat.Never.name
    db["items"][idx]["s_date"] = ""
    db["items"][idx]["n_date"] = ""
    if TodoStatus[item["status"]] is TodoStatus.Completed:
        # 只有当该事项状态为 Completed 时，才需要设置 dtime
        db["items"][idx]["dtime"] = now()
    update_db(db)
    ctx.exit()


def make_schedule(db: DB, i: int, every: str, start: Arrow, ctx: click.Context) -> None:
    """Set up a new schedule (repeat event)."""

    # 验证 start
    today = arrow.now()
    if start < today.floor("day"):  # floor("day") 返回本地时间当天零时零分
        click.echo("Error: Cannot start from a past day.")
        ctx.exit()

    # set "s_date"
    db["items"][i]["s_date"] = start.format(DateFormat)

    # set "dtime"
    # 一个事件只要设置了重复提醒，那么它的 dtime 就必须为零
    if db["items"][i]["dtime"]:
        db["items"][i]["dtime"] = 0

    # set "repeat"
    repeat = every.capitalize()
    if repeat not in (Repeat.Week.name, Repeat.Month.name, Repeat.Year.name):
        click.echo(f"Error: Cannot set '-every' to {every}")
        click.echo("Try 'todo repeat --help' to get more information.")
        ctx.exit()
    db["items"][i]["repeat"] = repeat

    # set "status" and "n_date"
    # 在本函数的开头已经验证过 start, 防止其小于今天。
    if start > today.ceil("day"):  # ceil("day") 返回本地时间当天最后一秒
        db["items"][i]["status"] = TodoStatus.Completed.name
        db["items"][i]["n_date"] = db["items"][i]["s_date"]
    if start.format(DateFormat) == today.format(DateFormat):
        db["items"][i]["status"] = TodoStatus.Incomplete.name
        n_date = shift_next_date(start, start, Repeat[repeat])
        db["items"][i]["n_date"] = n_date.format(DateFormat)


# 在调用本函数之前，要限定 repeat 的值的范围。
def shift_next_date(s_date: Arrow, n_date: Arrow, repeat: Repeat) -> Arrow:
    match repeat:
        case Repeat.Week:
            return n_date.shift(days=7)
        case Repeat.Month:
            if is_last_day(s_date):
                return n_date.shift(months=1).ceil("month")
            return n_date.shift(months=1)
        case Repeat.Year:
            if is_last_day(s_date):
                return n_date.shift(year=1).ceil("month")
            return n_date.shift(year=1)
        case _:
            raise ValueError(repeat)


def update_schedules(db: DB) -> None:
    today = arrow.now()
    u_date = today.format(DateFormat)
    if u_date == db["u_date"]:
        # 如果今天已经更新过，就不用更新了（每天只更新一次）
        return
    db["u_date"] = u_date
    for idx, item in enumerate(db["items"]):
        if today.format(DateFormat) == item["n_date"]:
            db["items"][idx]["status"] = TodoStatus.Incomplete.name
            s_date = arrow.get(item["s_date"])
            n_date = arrow.get(item["n_date"])
            n_date = shift_next_date(s_date, n_date, Repeat[item["repeat"]])
            db["items"][idx]["n_date"] = n_date.format(DateFormat)
    update_db(db)



import arrow
import click
from typing import cast

from simpletodo.model import (
    ErrMsg,
    Repeat,
    TodoStatus,
    new_todoitem,
    now,
)
from simpletodo.util import (
    db_path,
    ensure_db_file,
    load_db,
    print_donelist,
    print_repeatlist,
    print_todolist,
    split_db,
    stop_schedule,
    update_db,
    validate_n,
    make_schedule,
    update_schedules,
)
from . import (
    __version__,
    __package_name__,
)

CONTEXT_SETTINGS = dict(help_option_names=["-h", "--help"])


def check(ctx: click.Context, err: ErrMsg) -> None:
    """检查 err, 有错误则打印并终止程序，无错误则什么都不用做。"""
    if err:
        click.echo(f"Error: {err}")
        ctx.exit()


def show_where(ctx: click.Context, param, value):
    if not value or ctx.resilient_parsing:
        return
    click.echo(f"[todo] {__file__}")
    click.echo(f"[database] {db_path}")
    ctx.exit()


def dump(ctx: click.Context, param, value):
    if not value or ctx.resilient_parsing:
        return
    with open(db_path, "rb") as f:
        click.echo(f.read())
    ctx.exit()


@click.group(invoke_without_command=True)
@click.help_option("-h", "--help")
@click.version_option(
    __version__,
    "-V",
    "--version",
    package_name=__package_name__,
    message="%(prog)s version: %(version)s",
)
@click.option(
    "-w",
    "--where",
    is_flag=True,
    help="Show locations about simple-todo.",
    expose_value=False,
    callback=show_where,
)
@click.option(
    "-d",
    "--dump",
    is_flag=True,
    help="Dump out the database (a json file).",
    expose_value=False,
    callback=dump,
)
@click.option(
    "all",
    "-a",
    "--all",
    is_flag=True,
    help="Show all items (including 'Completed' and 'Schedule').",
)
@click.pass_context
def cli(ctx, all):
    """simple-todo: Yet another command line TODO tool (命令行TODO工具)

    Just run 'todo' (with no options and no command) to list all items.

    https://pypi.org/project/simpletodo/
    """
    if ctx.invoked_subcommand is None:
        db = load_db()
        update_schedules(db)
        if not db["items"]:
            click.echo("There's no todo item.")
            click.echo("Use 'todo add ...' to add a todo item.")
            click.echo("Use 'todo --help' to get more information.")
            ctx.exit()

        todo_list, done_list, repeat_list = split_db(db)
        print_todolist(todo_list, all)

        if all:
            print_donelist(done_list)
            print_repeatlist(repeat_list)

        print()


# 以上是主命令
############
# 以下是子命令


@cli.command(context_settings=CONTEXT_SETTINGS)
@click.argument("event", nargs=-1, required=True)
@click.pass_context
def add(ctx, event):
    """Add an event to the todo list.

    [EVENT] is a string describing a todo item.

    Example: todo add Buy more beer.
    """
    event = cast(tuple[str], event)
    subject = " ".join(event).strip()
    if not subject:
        click.echo(ctx.get_help())
        ctx.exit()

    db = load_db()
    db["items"].insert(0, new_todoitem(subject))
    update_db(db)
    ctx.exit()


@cli.command(context_settings=CONTEXT_SETTINGS)
@click.argument("n", nargs=1, type=click.INT)
@click.pass_context
def done(ctx, n):
    """Mark the N'th item as 'Completed'.

    Example: todo done 1
    """
    db = load_db()
    err = validate_n(db["items"], n)
    check(ctx, err)

    status = db["items"][n - 1]["status"]
    repeat = db["items"][n - 1]["repeat"]
    if TodoStatus[status] is TodoStatus.Completed:
        click.echo("Warning: It was in the completed-list, nothing changes.")
    db["items"][n - 1]["status"] = TodoStatus.Completed.name
    if Repeat[repeat] is Repeat.Never:
        # 如果设置了重复提醒，则不修改 dtime (让它的值保持为零)
        db["items"][n - 1]["dtime"] = now()
    update_db(db)
    ctx.exit()


@cli.command(context_settings=CONTEXT_SETTINGS)
@click.argument("n", nargs=1, type=click.INT)
@click.pass_context
def delete(ctx, n):
    """Delete the N'th item. (It will be removed, not marked as completed)

    Example: todo delete 2
    """
    db = load_db()
    err = validate_n(db["items"], n)
    check(ctx, err)
    del db["items"][n - 1]
    update_db(db)
    ctx.exit()


@cli.command(context_settings=CONTEXT_SETTINGS)
@click.pass_context
def clean(ctx):
    """Clear the completed list (delete all completed items)."""
    db = load_db()
    _, done_list, _ = split_db(db)
    for idx, _ in done_list:
        del db["items"][idx]
    update_db(db)
    ctx.exit()


@cli.command(context_settings=CONTEXT_SETTINGS)
@click.argument("n", nargs=1, type=click.INT)
@click.pass_context
def redo(ctx, n):
    """Mark the N'th item as 'Incomplete'.

    Example: todo redo 1
    """
    db = load_db()
    err = validate_n(db["items"], n)
    check(ctx, err)

    status = db["items"][n - 1]["status"]
    if TodoStatus[status] is TodoStatus.Incomplete:
        click.echo("Warning: It is in the incomplete-list, nothing changes.")

    db["items"][n - 1]["status"] = TodoStatus.Incomplete.name
    db["items"][n - 1]["ctime"] = now()
    db["items"][n - 1]["dtime"] = 0
    update_db(db)
    ctx.exit()


@cli.command(context_settings=CONTEXT_SETTINGS)
@click.argument("n", nargs=1, type=click.INT)
@click.option("every", "-every", "--every", help="Every 'week' or 'month' or 'year'.")
@click.option(
    "start",
    "-from",
    "--start-from",
    help="Example: -from 2021-04-01",
)
@click.option(
    "stop",
    "-stop",
    "--stop",
    is_flag=True,
    help="Stop repeating the event (删除指定事项的周期计划)",
)
@click.pass_context
def repeat(ctx, n, every, start: str, stop):
    """Set the N'th item to repeat every week/month/year.

    Example: todo repeat 1 -every month -from today
    """
    db = load_db()
    err = validate_n(db["items"], n)
    check(ctx, err)

    idx = n - 1
    item = db["items"][idx]

    # 优先处理 stop
    if stop:
        stop_schedule(db, idx, item, ctx)

    # 为了逻辑清晰，要求同时设置重复模式与起始时间
    if not every:
        click.echo("Error: Missing option '-every'.")
        click.echo("Try 'todo repeat --help' to get more information")
        ctx.exit()
    if not start:
        click.echo("Error: Missing option '-from'.")
        click.echo("Try 'todo repeat --help' to get more information")
        ctx.exit()

    today = arrow.now()
    match start.lower():
        case "today":
            s_date = today
        case "tomorrow":
            s_date = today.shift(days=1)
        case _:
            s_date = arrow.get(start)

    make_schedule(db, idx, every, s_date, ctx)
    update_db(db)
    ctx.exit()


@cli.command(context_settings=CONTEXT_SETTINGS)
@click.argument("args", type=(int, str))
@click.pass_context
def edit(ctx, args):
    """Edit the subject of an event.

    [ARGS] is a tuple[int, str].

    Example: todo edit 1 "Meet John on friday."
    """
    n, subject = args
    db = load_db()
    err = validate_n(db["items"], n)
    check(ctx, err)

    subject = subject.strip()
    if not subject:
        click.echo(ctx.get_help())
        ctx.exit()

    db["items"][n - 1]["event"] = subject
    update_db(db)
    ctx.exit()


# 初始化
ensure_db_file()

if __name__ == "__main__":
    cli(obj={})
